package tomracer.genetics;

import java.io.Serializable;
import java.text.DecimalFormat;
import java.util.Arrays;
import java.util.Random;
import java.util.Set;
import java.util.TreeSet;

public class Genome implements Serializable{
	/**
	 * 
	 */
	private static final long serialVersionUID = 1L;
	private double[] genes;
	private double minValue;
	private double maxValue;
	private static Random randomizer = Randomizer.getRandomizer();

	double fitness;
	Set<Integer> crossoverPoints;
	public Genome(double[] genes, double minValue, double maxValue, Set<Integer> crossoverPoints) {
		this.genes = genes.clone();
		this.minValue = minValue;
		this.maxValue = maxValue;
		fitness = -1;
		this.crossoverPoints = crossoverPoints;		
	}
	public Genome(double[] genes, double minValue, double maxValue) {
		this(genes,minValue,maxValue,null);
	}	
	public Genome(int size, double minValue, double maxValue) {
		this(new double[size],minValue,maxValue);
	}
	
	public Genome(int size, double minValue2, double maxValue2, Set<Integer> crossoverPoints2) {
		this(new double[size],minValue2,maxValue2,crossoverPoints2);
	}
	public Genome(Genome genome) {
		this.genes = genome.genes.clone();
		this.minValue = genome.getMinValue();
		this.maxValue = genome.getMaxValue();
		this.fitness = -1;
		this.crossoverPoints = genome.crossoverPoints;
	}
	public double[] getGenes() {
		return genes;
	}
	
	public void addNewCrossoverPoint(int point) {
		if (crossoverPoints == null) {
			crossoverPoints = new TreeSet<Integer>();
		}
		this.crossoverPoints.add(point);
	}
	
	public void mutate(double rate) {
		for (int i=0;i<genes.length;i++) {
			if (randomizer.nextDouble()<rate) {
				genes[i] = randomizer.nextDouble()*(maxValue-minValue) + minValue;
			}
		}
	}
	
	public void gaussMutate(double rate, double stdDev) {
		for (int i = 0; i < size(); i++) {
			if (randomizer.nextDouble()<rate) {
                genes[i] = genes[i] + (randomizer.nextGaussian() * stdDev);
                if (genes[i]>maxValue)
                	genes[i]=maxValue;
                if (genes[i]<minValue)
                	genes[i]=minValue;

			}
    	}
	}
	
	public void compositeMutate(double rate, double gaussProp, double stdDev) {
		for (int i = 0; i < size(); i++) {
			if (randomizer.nextDouble()<rate) {
				if (randomizer.nextDouble()<gaussProp) {
					genes[i] = genes[i] + (randomizer.nextGaussian() * stdDev);
				} else {
					genes[i] = randomizer.nextDouble()*(maxValue-minValue) + minValue;
				}
				if (genes[i]>maxValue)
                	genes[i]=maxValue;
                if (genes[i]<minValue)
                	genes[i]=minValue;

			}
    	}		
	}
	
	public static Genome getRandomGenome(int size, double minValue, double maxValue) {
		Genome result = new Genome(size,minValue,maxValue);
		result.mutate(1);
		return result;
	}
	public static Genome getRandomGenome(int size, double minValue, double maxValue, Set<Integer> crossoverPoints) {
		Genome result = new Genome(size,minValue,maxValue,crossoverPoints);
		result.mutate(1);
		return result;
	}
	public static Genome getZeroGenome(int size, double minValue, double maxValue) {
		double[] genes = new double[size];
		Arrays.fill(genes, (maxValue+minValue)/2.0);
		Genome result = new Genome(genes,minValue,maxValue);
		return result;
	}
	public static Genome getZeroGenome(int size, double minValue, double maxValue, Set<Integer> crossoverPoints) {
		double[] genes = new double[size];
		Arrays.fill(genes, (maxValue+minValue)/2.0);
		Genome result = new Genome(genes,minValue,maxValue, crossoverPoints);
		return result;
	}

	
	public static Genome crossover(Genome g1, Genome g2) {
		int size = g1.size();
		int p1 = randomizer.nextInt(size);
		int p2 = randomizer.nextInt(size+p1);
		if (p1 >p2) {
			int dummy = p1;
			p1 = p2;
			p2 = dummy;
		}
		double[] newGenes = g1.getGenes().clone();
		for (int i=p1;i<p2;i++) {
			newGenes[i%size] = g2.getGene(i%size);
		}
		return new Genome(newGenes,g1.getMinValue(),g1.getMaxValue(),g1.getCrossoverPoints());
	}
	
	public static Genome crossoverUsingPoints(Genome g1, Genome g2) {
//		Set<Integer> crossoverPoints = g1.crossoverPoints;
//		if (crossoverPoints == null) {
//			System.err.println("trying crossover point crossover but no points defined");
//			return crossover(g1,g2);
//		}
		Genome[] g = {g1,g2};
		double[] newGenes = new double[g1.size()];
		int current = randomizer.nextInt(2);
		for (int i=0;i<g1.size();i++) {
//			if (crossoverPoints.contains(i)) {
				current = randomizer.nextInt(2);
//			}
			newGenes[i] = g[current].getGene(i);
		}
		return new Genome(newGenes,g1.getMinValue(),g1.getMaxValue(), g1.getCrossoverPoints());
	}
	
	public double getMaxValue() {
		return maxValue;
	}
	public double getMinValue() {
		return minValue;
	}
	public double getGene(int i) {
		return genes[i%size()];
	}
	public int size() {
		return genes.length;
	}
	
	public void setFitness(double fitness) {
		this.fitness = fitness;
	}
	
	public double getFitness() {
		return fitness;
	}
	
	
	
	public String toString() {
		String toString = arrayToString(genes);
		return toString;
	}
	public static String arrayToString(double[] array) {
		DecimalFormat df = new DecimalFormat();
		df.setMaximumFractionDigits(2);
		df.setMinimumFractionDigits(2);
		StringBuffer sb = new StringBuffer("{"+df.format(array[0]));

		for (int i = 1; i < array.length; i++) {
			sb.append(","+df.format(array[i]));
		}
		sb.append("}");
		String toString = sb.toString();
		return toString;
	}
	public static void main(String[] args) {
		double[] genes1 = {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
		double[] genes2 = {2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2};
		Genome genome1 = new Genome(genes1,-2,2);
		Genome genome2 = new Genome(genes2,-2,2);
		System.out.println(genome1);
		System.out.println(genome2);
		genome1.gaussMutate(0.4,0.3);
		System.out.println(genome1);
		System.out.println(Genome.crossover(genome1, genome2));
	}
	public void setGene(int i, double gene) {
		genes[i] = gene;
	}
	public Set<Integer> getCrossoverPoints() {
		// TODO Auto-generated method stub
		return crossoverPoints;
	}
	public void setGenes(double[] evolvedGenes) {
		this.genes = evolvedGenes;
	}
}